@@ -5,7 +5,8 @@
 
 // [[export]]
 SEXP agg_jpeg_c(SEXP file, SEXP width, SEXP height, SEXP pointsize, SEXP bg, 
-               SEXP res, SEXP quality, SEXP smoothing, SEXP method) {
+                SEXP res, SEXP scaling, SEXP quality, SEXP smoothing, 
+                SEXP method) {
   int bgCol = RGBpar(bg, 0);
   
   BEGIN_CPP
@@ -16,6 +17,7 @@ SEXP agg_jpeg_c(SEXP file, SEXP width, SEXP height, SEXP pointsize, SEXP bg,
       REAL(pointsize)[0], 
       bgCol,
       REAL(res)[0],
+      REAL(scaling)[0],
       INTEGER(quality)[0],
       INTEGER(smoothing)[0],
       INTEGER(method)[0]
